#ifndef EIGENVALUEDECOMP_H_
#define EIGENVALUEDECOMP_H_

#include <math.h>
//#include "nrutil.h"

// library includings
#include "num_linalg.h"

// own includings
#include "TransformMatrices.h"

#define ROTATE(a,i,j,k,l) g=a(i,j); h=a(k,l); a(i,j)=g-s*(h+g*tau); a(k,l)=h+s*(g-h*tau);

namespace Basics3D
{
	// nrc_2 p.491
	// Computes all eigenvalues and eigenvectors of a real symmetric
	// matrix a[1..n][1..n]. On output, elements of a above the
	// diagonal are destroyed. d[1..n] returns the eigenvalues of a.
	// v[1..n][1..n] is a matrix whose columns contain, on output,
	// the normalized eigenvectors of a.
	// nrot returns the number of Jacobi rotations that were required.
	template <class MatrixType>
	void jacobi(MatrixType &in_a,
		          std::vector<double> &out_d,
	            MatrixType &out_v,
							int *out_nrot)
	{
		if (in_a.size1() != in_a.size2())
		{
			throw "The matrix to decompose should be quadric";
			return;
		}

		if (in_a.size1() != out_v.size1())
		{
			throw "The matrix of eigen vectors should have the same size\nas the matrix to decompose";
			return;
		}

		if (out_v.size1() != out_v.size2())
		{
			throw "The matrix of eigen vectors should be quadric";
			return;
		}

		int n = in_a.size1();	// size of input matrix and matrix of eigen vectors
		//std::cout << std::endl << "n = " << n;

		Vtl::NumAlg::Lin::ublas::vector<double> b(n);

		// This vector will accumulate terms of the form tapq as in equation (11.1.14).
		Vtl::NumAlg::Lin::ublas::vector<double> z(n);
		
		//Initialize to the identity matrix.
		C_TransformMatrices::Identity(out_v);

		// Initialize b and d to the diagonal of a.
		for (int ip = 0; ip < n; ip++)
		{ 
			b[ip] = out_d[ip] = in_a(ip, ip);
			z[ip] = 0.0; 
		}

		*out_nrot = 0;

		for (int i = 1; i <= 100; i++)
		{
			// Sum off-diagonal elements.
			double sm = 0.0;
			for (int ip = 0; ip < n - 1; ip++)
			{ 
				for (int iq = ip + 1; iq < n; iq++)
				{
					sm += fabs(in_a(ip, iq));
				}
			}
			std::cout << std::endl << "Sum off-diagonal elements = " << sm;

			// The normal return, which relies on quadratic convergence to machine underflow.
			if (sm < 0.0000000000000001)
			{ 
				return;
			}

			double tresh;
			// ...on the first three sweeps.
			if (i < 4)
			{
				tresh = 0.2 * sm / (n * n); 
			}
			// ...thereafter.
			else
			{
				tresh = 0.0;
			}

			for (int ip = 0; ip < n - 1; ip++)
			{
				//std::cout << std::endl << "ip = " << ip << "       ";
				for (int iq = ip + 1; iq < n; iq++)
				{
					//std::cout << "iq = " << iq << "  ";
					double g = 100.0 * fabs(in_a(ip, iq));

					// 	After four sweeps, skip the rotation if the off-diagonal element is small.
					if (i > 4 && (fabs(out_d[ip]) + g == fabs(out_d[ip]))
						  && (fabs(out_d[iq]) + g ==  fabs(out_d[iq])))
					{
						in_a(ip, iq) = 0.0;
					}
					else if (fabs(in_a(ip, iq)) > tresh)
					{
						double h = out_d[iq] - out_d[ip];

						double t;
						double theta;
						if (fabs(h) + g == fabs(h))
						{
							t = in_a(ip, iq) / h; //t = 1/(2*teta)
						}
						else 
						{
							//Equation (11.1.10).
							theta = 0.5 * h / in_a(ip, iq);		
							t = 1.0 / (fabs(theta) + sqrt(1.0 + theta * theta));
							if (theta < 0.0)
							{
								t = -t;
							}
						}

						double c = 1.0 / sqrt(1 + t * t);
						double s = t * c;
						double tau = s / (1.0 + c);
						h = t * in_a(ip, iq);

						z[ip] -= h;
						z[iq] += h;
						out_d[ip] -= h;
						out_d[iq] += h;
						in_a(ip, iq) = 0.0;

						// Case of rotations 0 <= j < p - 1.
						for (int j = 1; j < ip; j++)
						{ 
							ROTATE(in_a, j, ip, j, iq)
						}
						// Case of rotations p - 1 < j < q - 1.
						for (int j = ip + 1; j < iq; j++)
						{ 
							ROTATE(in_a, ip, j, j, iq)
						}
						// Case of rotations q - 1 < j < n.
						for (int j = iq + 1; j < n; j++)
						{ 
							ROTATE(in_a, ip, j, iq, j)
						}
						for (int j = 0; j < n; j++)
						{
							ROTATE(out_v, j, ip, j, iq)
						}

						++(*out_nrot);
					}
				}
			}

			for (int ip = 0; ip < n; ip++)
			{
				b[ip] += z[ip];
				out_d[ip] = b[ip];		//Update d with the sum of tapq,
				z[ip] = 0.0;					//and reinitialize z.
			}
		}

		throw "Too many iterations in routine jacobi";
		return;
	}

}


#endif // EIGENVALUEDECOMP_H_
