#ifndef FUNDAMENTAL_MATRIX_H
#define FUNDAMENTAL_MATRIX_H

#include <iostream>
#include <vector>
#include <math.h>

// library includings
#include "nummet.h"
#include "num_linalg.h"

// own includings
#include "MatrixOperations.h"
#include "Point2D.h"
#include "Points2DSet.h"
#include "Points2DParamSet.h"
#include "SolveSystem.h"
#include "Ransac.h"
#include "RecompMatrix.h"
#include "Model.h"
#include "ImagesSet.h"

namespace Basics3D
{
	void PrintAndSave
		      (const char *in_name,
           const std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> &in_Matrix,
					 const int i,
					 const int j);

	void DrawEpipolarLines(const T_ImagesSet<AImageMem8> &in_img,
		                     const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_Matrix,
						             const std::vector<T_Points2DSet<C_Point2D>> &in_PointsSet,
						             const int i,
						             const int n);

	class C_FundamentalMatrix
	{
	private:
		static void SevenPointsAlg
			                (std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> &out,
										   const Vtl::NumAlg::Lin::ublas::matrix<double> &in_System);

		static void EightPointsAlg(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out,
										           const Vtl::NumAlg::Lin::ublas::matrix<double> &in_System);
	public:
		// Returns minimal number of points pairs requared for computation of fund. matrix
		static int SampleSize();
		
		// Returns the distance threshold
		// in_sigma is inaccuracy of feature points detector
		static double DistThresh(double in_sigma);

		// Creates matrix A of the system Af = 0, 
		// where f is a vector containing elements of the fundamental matrix,
		//       strings of A: (x'x, x'y, x', y'x, y'y, y', x, y, 1)
		//       (x, y), (x', y') - corresponding points on two images
		static void CreateSystem(Vtl::NumAlg::Lin::ublas::matrix<double> &out_M,
														 const T_Points2DSet<C_Point2D> &in_Points);

		// Creates system to find FM from homography matrices
		static void CreateSystem
			(Vtl::NumAlg::Lin::ublas::matrix<double> &out_M,
			 const std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> &in_Matrices);

		// Finds fundamental matrix
		// out is an array of resulting matrices (1 or 3)
		// &in_System is a matrix of system to solve obtained by CreateSystem
		static void Find(std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>>& out,
										 const Vtl::NumAlg::Lin::ublas::matrix<double> &in_System);

		static double Dist(const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 1> &in_p1,
											 const C_Point2D &in_p2);

		static double MistFunc(const Vtl::NumAlg::Lin::ublas::vector<double> &in_matrix,
			                     const C_Point2D &in_p1,
			                     const C_Point2D &in_p2);

		static void DerivativeOfMistFunc(Vtl::NumAlg::Lin::ublas::vector<double> &out,
			                               const Vtl::NumAlg::Lin::ublas::vector<double> &in_matrix,
			                               const C_Point2D &in_p1,
			                               const C_Point2D &in_p2);

		static void NormMatrix(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out_M,
			                     Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_M,
													const double in_err);

		static void FromHomographies
			        (std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> &out_FM,
			         const std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> &in_H);

		static void ApplyRansac
			         (std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> &out_matrix,
		            std::vector<T_Points2DSet<C_Point2D>> &out_inliers,
								std::vector<T_Points2DSet<C_Point2D>> &in_pointsSet,
		            const T_ImagesSet<AImageMem8> &in_img,
		            const int in_i,
		            const double in_sigma);

		static void RecomputeWithAllPoints
			         (std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> &out_matrix,
		            std::vector<T_Points2DSet<C_Point2D>> &out_inliers,
								std::vector<C_Points2DParamSet> &in_pointsParamSet,
								const T_ImagesSet<AImageMem8> &in_img,
		            const int in_i,
								const double in_sigma,
								const double in_t);
	};
}
#endif //FUNDAMENTAL_MATRIX_H
