#ifndef LMMINIMIZATION_H_
#define LMMINIMIZATION_H_
#include <math.h>
#include <vector>
#include "nrutil.h"

void SWAP(float a, float b)
{
	float swap = a;
	a = b;
	b = swap;
}
//#define SWAP(a,b) {swap=(a);(a)=(b);(b)=swap;}

/////////////////////////////////////////////////////////////////////////////////////////
// Linear equation solution by Gauss-Jordan elimination, equation (2.1.1) above.
// a[1..n][1..n] is the input matrix.
// b[1..n][1..m] is input containing the m right-hand side vectors.
// On output, a is replaced by its matrix inverse, and b is replaced by the corresponding
// set of solution vectors.
void gaussj(float **a,
						int n,
						float **b,
						int m)
{
	int *indxc, *indxr, *ipiv;
	int i, icol, irow, j, k, l, ll;
	float big, dum, pivinv/*, temp*/;

	// The integer arrays ipiv, indxr, and indxc are used for bookkeeping on the pivoting.
	indxc = ivector(1, n); 
	indxr = ivector(1, n); 
	ipiv = ivector(1, n);

	for (j = 0; j < n; j++)		//(j = 1; j <= n; j++)
	{
		ipiv[j] = 0;
	}

	// This is the main loop over the columns to be reduced.
	for (i = 0; i < n; i++)		//(i = 1; i <= n; i++)
	{ 
		big = 0.0;

		// This is the outer loop of the search for a pivot element.
		for (j = 0; j < n; j++)		//(j = 1; j <= n; j++) 
			if (ipiv[j] != 1) 
				for (k = 0; k < n; k++)		// (k = 1; k <= n; k++)
				{
					if (ipiv[k] == 0) {
						if (fabs(a[j][k]) >= big) {
							big=fabs(a[j][k]);
							irow=j;
							icol=k;
						}
					} else if (ipiv[k] > 1) nrerror("gaussj: Singular Matrix-1");
				}
				++(ipiv[icol]);

				// We now have the pivot element, so we interchange rows, if needed, to put
				// the pivot element on the diagonal. The columns are not physically
				// interchanged, only relabeled: indxc[i], the column of the ith pivot element,
				// is the ith column that is reduced, while indxr[i] is the row in which that
				// pivot element was originally located. If indxr[i] 6=	indxc[i] there is an
				// implied column interchange. With this form of bookkeeping, the solution b's
				// will end up in the correct order, and the inverse matrix will be scrambled
				// by columns.
					if (irow != icol) {
						for (l = 0; l < n; l++)/*(l=1; l<=n; l++)*/ SWAP(a[irow][l], a[icol][l]);
						for (l = 0; l < m; l++) /*(l=1;l<=m;l++)*/SWAP(b[irow][l], b[icol][l]);
					}

					// We are now ready to divide the pivot row by the pivot element,
					// located at irow and icol.
					indxr[i]=irow; 
						indxc[i]=icol;  
						if (a[icol][icol] == 0.0) nrerror("gaussj: Singular Matrix-2");
					pivinv=1.0/a[icol][icol];
					a[icol][icol]=1.0;
					for (l = 0; l < n; l++) /*(l=1;l<=n;l++)*/a[icol][l] *= pivinv;
					for (l = 0; l < m; l++) /*(l=1;l<=m;l++)*/b[icol][l] *= pivinv;

					// Next, we reduce the rows ...except for the pivot one, of course.
					for (ll = 0; ll < n; ll++)  //(ll=1;ll<=n;ll++) 
						if (ll != icol) { 
							dum=a[ll][icol];
					a[ll][icol]=0.0;
					for (l = 0; l < n; l++) /*(l=1;l<=n;l++)*/a[ll][l] -= a[icol][l]*dum;
					for (l = 0; l < m; l++) /*(l=1;l<=m;l++)*/b[ll][l] -= b[icol][l]*dum;
						}
	}

	// This is the end of the main loop over columns of the reduction.
	// It only remains to unscramble the solution in view of the column interchanges. We do
	// this by interchanging pairs of columns in the reverse order that the permutation was
	// built up.
		for (l = n-1; l >= 0 ; l--)	//(l=n;l>=1;l--)
		{
			if (indxr[l] != indxc[l])
				for (k = 0; k < n; k++)		//(k=1;k<=n;k++)
					SWAP(a[k][indxr[l]],a[k][indxc[l]]);
		}
		// And we are done.
		free_ivector(ipiv,1,n);
		free_ivector(indxr,1,n);
		free_ivector(indxc,1,n);
}

/////////////////////////////////////////////////////////////////////////////////////////
// Expand in storage the covariance matrix covar, so as to take into account parameters that are
// being held fixed. (For the latter, return zero covariances.)
void covsrt(float **covar, 
						int ma,
						int ia[],
						int mfit)
{
	int i,j,k;
	/*float swap;*/

	for (i = mfit; i < ma; i++)		// (i = mfit + 1; i <= ma; i++)
	{
		for (j = 0; j < i; j ++)		// (j = 1; j <= i; j ++)
		{
			covar[i][j] = covar[j][i] = 0.0;
		}

		k = mfit;

		for (j = ma - 1; j >= 0; j--)		//(j = ma; j >= 1; j--)
		{
			if (ia[j])
			{
				for (i = 0; i < ma; i++)		//(i = 1; i <= ma; i++)
				{
					SWAP(covar[i][k], covar[i][j]);
				}

				for (i = 0; i < ma; i++)		//(i = 1; i <= ma; i++)
				{
					SWAP(covar[k][i], covar[j][i]);
				}

				k--;
			}
		}
	}
}

/////////////////////////////////////////////////////////////////////////////////////////
// Levenberg-Marquardt method, attempting to reduce the value chisq of a fit between
// a set of data points x[1..ndata], y[1..ndata] with individual standard deviations
// sig[1..ndata], and a nonlinear function dependent on ma coeffcients a[1..ma]. 
// The input array ia[1..ma] indicates by nonzero entries those components of a that
// should be fitted for, and by zero entries those components that should be held fixed
// at their input values.
// The program returns current best-fit values for the parameters a[1..ma], and chisq.
// The arrays covar[1..ma][1..ma], alpha[1..ma][1..ma] are used as working space during
// most iterations. 
// Supply a routine funcs(x, a, yfit, dyda, ma) that evaluates the fitting function
// yfit, and its derivatives dyda[1..ma] with respect to the ftting parameters a at x.
// On the first call provide an initial guess for the parameters a, and set alamda < 0
// for initialization (which then sets alamda = 0.001). If a step succeeds chisq becomes
// smaller and alamda decreases by a factor of 10. If a step fails alamda grows by
// a factor of 10. You must call this routine repeatedly until convergence is achieved.
// Then, make one final call with alamda = 0, so that covar[1..ma][1..ma] returns the 
// covariance matrix, and alpha the curvature matrix.
// (Parameters held fixed will return zero covariances.)
void mrqmin(float x[],
						/*float y[],*/
						float y,
						float sig/*[]*/,
						int ndata,
						float a[],
						int ia[],
            int ma,
						float **covar,
						float **alpha,
						float *chisq,
						void (*funcs)(float [],
						              float [],
													float *,
													float [],
													int),
						float *alamda)
{
	void covsrt(float **covar, int ma, int ia[], int mfit);
	void gaussj(float **a, int n, float **b, int m);
	void mrqcof(float x[], float y/*[]*/, float sig/*[]*/, int ndata, float a[],
		int ia[], int ma, float **alpha, float beta[], float *chisq,
		void (*funcs)(float[], float [], float *, float [], int));

	int j,k,l;
	static int mfit;
	static float ochisq, *atry, *beta, *da, **oneda;

	// Initialization
	if (*alamda < 0.0)
	{ 
		atry = vector(1, ma);
		beta = vector(1, ma);
		da = vector(1, ma);

		for (mfit = 0, j = 0; j < ma; j++)		//(mfit = 0, j = 1; j <= ma; j++)
		{
			if (ia[j])
			{
				mfit++;
			}
		}

		oneda = matrix(0, mfit-1, 0, 0);
		*alamda = 0.001;
		mrqcof(x, y, sig, ndata, a, ia, ma, alpha, beta, chisq, funcs);
		ochisq = (*chisq);

		for (j = 0; j < ma; j++)		//(j = 1; j <= ma; j++)
		{
			atry[j]=a[j];
		}
	}

	// Alter linearized fitting matrix, by augmenting diagonal elements.
	for (j = 0; j < mfit; j++)	//(j = 1; j <= mfit; j++)
	{ 
		for (k = 0; k < mfit; k++)		//(k = 1; k <= mfit; k++)
		{
			covar[j][k] = alpha[j][k]; 
		}

		covar[j][j] = alpha[j][j] * (1.0 + (*alamda));
	  oneda[j][0] = beta[j];		//oneda[j][1] = beta[j];
	}

	// Matrix solution.
	gaussj(covar, mfit, oneda, 1);

	for (j = 0; j < mfit; j++)		//(j = 1; j <= mfit; j++)
	{
		da[j] = oneda[j][1];
	}

	// Once converged, evaluate covariance matrix.
	if (*alamda == 0.0)
	{ 
		covsrt(covar, ma, ia, mfit);
	  covsrt(alpha, ma, ia, mfit);				//Spread out alpha to its full size too.
		free_matrix(oneda, 0, mfit-1, 0, 0);		//(oneda, 1, mfit, 1, 1);
		free_vector(da, 0, ma-1);								//(da, 1, ma);
		free_vector(beta, 0, ma-1);							//(beta, 1, ma);
		free_vector(atry, 0, ma-1);							//(atry, 1, ma);
		return;
	}

	// Did the trial succeed?
	for (j = 0, l = 0; l < ma; l++)		//(j = 0, l = 1; l <= ma; l++) 
	{
		if (ia[l])
		{
			atry[l] = a[l] + da[++j];
		}
	}

	mrqcof(x, y, sig, ndata, atry, ia, ma, covar, da, chisq, funcs);

	// Success, accept the new solution.
	if (*chisq < ochisq)
	{ 
		*alamda *= 0.1;
		ochisq = (*chisq);

		for (j = 0; j < mfit; j++)		//(j = 1; j <= mfit; j++)
		{
		  for (k = 0; k < mfit; k++)		//(k = 1; k <= mfit; k++)
			{
				alpha[j][k] = covar[j][k];
			}

		  beta[j] = da[j];
	  }

	  for (l = 0; l < ma; l++)		//(l = 1; l <= ma; l++)
		{
			a[l]=atry[l];
		}
	} 
  // Failure, increase alamda and return.
	else
	{ 
		*alamda *= 10.0;
	  *chisq=ochisq;
	}
}

/////////////////////////////////////////////////////////////////////////////////////////
// Used by mrqmin to evaluate the linearized fitting matrix alpha, and vector beta
// as in (15.5.8), and calculate chisq.
void mrqcof(float x[],
						/*float y[],*/
						float y,
						float sig/*[]*/,
						int ndata,
						float a[],
						int ia[],
						int ma,
						float **alpha,
						float beta[],
						float *chisq,
						void (*funcs)(float [],
						              float [],
													float *,
													float [],
													int))
						
{
	int i, j, k, l, m, mfit = 0;
	float ymod, wt, sig2i, dy, *dyda;

	dyda = vector(1, ma);

	for (j = 0; j < ma; j++)		//(j = 1; j <= ma; j++)
	{
		if (ia[j]) 
		{
			mfit++;
		}
	}
	
	// Initialize (symmetric) alpha, beta.
	for (j = 0; j < mfit; j++)	// (j = 1; j <= mfit; j++)
	{ 
		for (k = 0; k < j; k++)  // (k = 1; k <= j; k++)
		{
			alpha[j][k] = 0.0;
		}

		beta[j]=0.0;
	}

	*chisq = 0.0;

	// Summation loop over all data.
	for (i = 0; i < ndata; i+=4/*+*/)		//(i = 1; i <= ndata; i+=4/*+*/)
	{ 
		(*funcs)(x + i /*[i]*/, a, &ymod, dyda, ma);
		sig2i = 1.0 / (sig/*[i]*/ * sig/*[i]*/);
		dy = y/*[i]*/ - ymod;

		for (j = 0, l = 0; l < ma; l++)		//(j = 0, l = 1; l <= ma; l++)
		{
			if (ia[l])
			{
				wt = dyda[l] * sig2i;

				for (j++, k=0 ,m = 0; m < l; m++)		//(j++, k=0 ,m = 1; m <= l; m++)
				{
					if (ia[m])
					{
						alpha[j][++k] += wt * dyda[m];
					}
				}

				beta[j] += dy * wt;
			}
		}

		// And find chisq.
		*chisq += dy * dy * sig2i; 
	}

	// Fill in the symmetric side.
	for (j = 1; j < mfit; j++)	// (j = 2; j <= mfit; j++)
	{
		for (k = 0; k < j; k++)		// (k = 1; k <= j; k++)
		{
			alpha[k][j] = alpha[j][k];
		}
	}

	free_vector(dyda, 1, ma);
}

/////////////////////////////////////////////////////////////////////////////////////////
//y(x; a) is the sum of na/3 Gaussians (15.5.16). The amplitude, center, and width of the
//Gaussians are stored in consecutive locations of a: a[i] = Bk, a[i+1] = Ek, a[i+2] = Gk,
//k = 1,...,na/3. The dimensions of the arrays are a[1..na], dyda[1..na].
//void fgauss(float x, float a[], float *y, float dyda[], int na)
//{
//	int i;
//	float fac,ex,arg;
//	*y=0.0;
//	for (i=1;i<=na-1;i+=3) {
//		arg=(x-a[i+1])/a[i+2];
//		ex=exp(-arg*arg);
//		fac=a[i]*ex*2.0*arg;
//		*y += a[i]*ex;
//		dyda[i]=ex;
//		dyda[i+1]=fac/a[i+2];
//		dyda[i+2]=fac*arg/a[i+2];
//	}
//}

void FToMinimaze(float x[], float a[], float *y, float dyda[], int ma)
{
	float x1 = x[0];
	float y1 = x[1];
	float x2 = x[2];
	float y2 = x[3];

	dyda[0] = x1 * x2;
	dyda[1] = x2 * y1;
	dyda[2] = x2;
	dyda[3] = x1 * y2;
	dyda[4] = y1 * y2;
	dyda[5] = y2;
	dyda[6] = x1;
	dyda[7] = y1;
	dyda[8] = 1;

	*y = x1 * x2 * a[0] + x2 * y1 * a[1] + x2 * a[2] + x1 * y2 * a[3] +
		   y1 * y2 * a[4] + y2 * a[5] + x1 * a[6] + y1 * a[7] + a[8];
}
#endif //LMMINIMIZATION_H_
