#ifndef MATRIXOPERATIONS_H_
#define MATRIXOPERATIONS_H_

#include <iostream>
#include <math.h>
#include <fstream>
#include <vector>

// library includings
#include "nummet.h"
#include "num_linalg.h"
#include "Sort.h"

namespace Basics3D
{
	class T_MatrixOperations 
	{
	public:
		
		template <class MatrixType>
		static void Normalize(MatrixType &out_M,
													const MatrixType &in_M,
													const MatrixType &T1_1,
													const MatrixType &T)
		{
			out_M = boost::numeric::ublas::prod(T1_1, in_M);
			out_M = boost::numeric::ublas::prod(out_M, T);
		}

		/////////////////////////////////////////////////////////////////////////////////////
		template <class MatrixType>
		static void TransformIntoNormCoord(std::vector<MatrixType> &out_matrix,
			                                 const int in_w,
			                                 const int in_h,
																			 const double in_sigma)
		{
			double d = sqrt( double(in_w * in_w + in_h * in_h));

			Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> T;
			double err = C_TransformMatrices::Normalization(T, in_w / 2.0, in_h / 2.0,
				                                              d / 2.0, in_sigma);
			Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> T_1;
			C_TransformMatrices::NormalizationInverse(T_1, in_w / 2.0, in_h / 2.0, 
				                                        d / 2.0, in_sigma);
			
			for (int j = 0; j < int(out_matrix.size()); j++)
			{
				T_MatrixOperations::Normalize(out_matrix[j], out_matrix[j], T, T_1);
			}
		}

		/////////////////////////////////////////////////////////////////////////////////////
		template <class MatrixType>
		static void TransformIntoPixelCoord(std::vector<MatrixType> &out_matrix,
			                                  const int in_w,
			                                  const int in_h,
																			  const double in_sigma)
		{
			double d = sqrt(double(in_w * in_w + in_h * in_h));
					
			Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> T;
			double err = C_TransformMatrices::Normalization(T, in_w / 2.0, in_h / 2.0,
				                                              d / 2.0, in_sigma);
			Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> TT;
			TT = boost::numeric::ublas::trans(T);

			for (int i = 0; i < (int)out_matrix.size(); i++)
			{
				T_MatrixOperations::Normalize(out_matrix[i], out_matrix[i], TT, T);
			}
		}

		/////////////////////////////////////////////////////////////////////////////////////
		// Sort columns in order given by indexes k[]
		template <class MatrixType>
		static void SortColumns(MatrixType &out_M,
			                      const MatrixType &in_M,
														const SortS<double> in_k[])
		{
			int m = in_M.size1();
			int n = in_M.size2();

			// checking of sizes
			if (out_M.size1() != m || out_M.size2() != n)
			{
				std::ostringstream os;
				os << "Function T_MatrixOperations::SortColumns: input and output "
					<< "matrices should have the same size";
				std::string str = os.str();
				throw str.c_str();
			}

			/*if (in_k.size() != n)
			{
				std::ostringstream os;
				os << "Function T_MatrixOperations::SortColumns: number of indexes "
					<< "and number of columns should be the same";
				std::string str = os.str();
				throw str.c_str();
			}*/

			std::cout << std::endl << std::endl << "in_M" << std::endl;
			T_MatrixOperations::Print(in_M);

			MatrixType tmp;

			// Sorting
			for (int i = 0; i < m; i++)
			{
				for (int j = 0; j < n; j++)
				{
					tmp(i, j) = in_M(i, n - in_k[j].index - 1);
				}
			}
			out_M = tmp;

			// Printing
			std::cout << std::endl << std::endl << "k" << std::endl;
			for (int i = 0; i < n; i++)
			{
				std::cout << in_k[i].index << "  ";
			}
			std::cout << std::endl << std::endl << "out_M" << std::endl;
			T_MatrixOperations::Print(out_M);
		}

		///////////////////////////////////////////////////////////////////////////////////////
		template <class MatrixType>
		static void Scale(MatrixType &out_M,
			                const MatrixType &in_M)
		{
			out_M.resize(in_M.size1(), in_M.size2(), false);

			double c = in_M(in_M.size1() - 1, in_M.size2() - 1);

			for (int i = 0; i < (int)in_M.size1(); i++)
			{
				for (int j = 0; j < (int)in_M.size2(); j++)
				{
					out_M(i, j) = in_M(i, j) / c;
				}
			}
		}

		///////////////////////////////////////////////////////////////////////////////////////
		template <class MatrixType>
		static void Print(const MatrixType &in_M)
		{
			for (int i = 0; i < (int)in_M.size1()/*m_m*/; i++)
			{
				if (i != 0) std::cout << "\n";

				for (int j = 0; j < (int)in_M.size2()/*m_n*/; j++)
				{
					/*if (abs(m_Ar[i * m_n + j]) < 0.000001)
					{
						std::cout.width(10);
						std::cout << 0;
					}
					else
					{*/
						std::cout.precision(3);
						std::cout.width(10);
						std::cout << in_M(i, j);//m_Ar[i * size2()/*m_n*/ + j];
					/*}*/
				}
			}
		}
		///////////////////////////////////////////////////////////////////////////////////////
		template <class MatrixType>
		static void Save (const MatrixType &in_M,
											const char *filename)
		{
			std::ofstream outFile;
			outFile.open(filename);

			for (int i = 0; i < (int)in_M.size1()/*m_m*/; i++)
			{
				for (int j = 0; j < (int)in_M.size2()/*m_n*/; j++)
				{
					outFile << in_M(i, j)/*m_Ar[i * m_n + j]*/ << " ";
				}
			}

			outFile.close();
		}

	};
}
#endif // MATRIXOPERATIONS_H_
