#ifndef POINTSSET_H_
#define POINTSSET_H_

#include <vector>
#include <fstream>
#include <math.h>

// own includings
#include "TransformMatrices.h"
#include "MatrixOperations.h"
#include "Point2D.h"

namespace Basics3D
{
	template<class Point>
	class T_Points2DSet
	{
	private:
		std::vector<Point> m_Points;
    int m_groupSize;
		std::vector<double> m_MeanX;
		std::vector<double> m_MeanY;
		std::vector<double> m_MeanDev;

	private:
		void ClearMeans()
		{
			for (int i = 0; i < m_groupSize; i++)
			{
				m_MeanX[i] = m_MeanY[i] = m_MeanDev[i] = -1;
			}
		}

		/////////////////////////////////////////////////////////////////////////////////////

	public:
		/////////////////////////////////////////////////////////////////////////////////////
		double MeanX(int i)
		{
			if (i >= m_groupSize)
			{
				throw "\nIndex is to be less than group size";
				return -1;
			}

			if (m_MeanX[i] == -1)
			{
				m_MeanX[i] = 0;

				for (int j = 0; j < Size(); j++)
				{
					m_MeanX[i] += m_Points[m_groupSize * j + i].X();
				}

				m_MeanX[i] /= Size();
			}

			return m_MeanX[i];
		}

		/////////////////////////////////////////////////////////////////////////////////////
		double MeanY(int i)
		{
			if (i >= m_groupSize)
			{
				throw "\nIndex is to be less than group size";
				return -1;
			}

			if (m_MeanY[i] == -1)
			{
				m_MeanY[i] = 0;

				for (int j = 0; j < Size(); j++)
				{
					m_MeanY[i] += m_Points[m_groupSize * j + i].Y();
				}

				m_MeanY[i] /= Size();
			}

			return m_MeanY[i];
		}

    /////////////////////////////////////////////////////////////////////////////////////
		double MeanDev(int i)
		{
			if (i >= m_groupSize)
			{
				throw "\nIndex is to be less than group size";
				return -1;
			}

			if (m_MeanDev[i] < 0)
			{
				m_MeanDev[i] = 0;

				for (int j = 0; j < Size(); j++)
				{
					C_Point2D P = m_Points[m_groupSize * j + i];
					m_MeanDev[i] += sqrt((P.X() - MeanX(i)) * (P.X() - MeanX(i)) +
															 (P.Y() - MeanY(i)) * (P.Y() - MeanY(i)));
				}

				m_MeanDev[i] /= Size();
			}
			return m_MeanDev[i];
		}

		/////////////////////////////////////////////////////////////////////////////////////
		T_Points2DSet(int in_groupSize = 1)
		{
			InitGroupSize(in_groupSize);
		}
			
		/////////////////////////////////////////////////////////////////////////////////////////
		T_Points2DSet<Point> operator = (const T_Points2DSet<C_Point2D> &in_Set)
		{
		  m_groupSize = in_Set.GroupSize();

			m_MeanX.resize(m_groupSize);
			m_MeanY.resize(m_groupSize);
			m_MeanDev.resize(m_groupSize);

			ClearMeans();

			m_Points.clear();

			int size = in_Set.Size() * in_Set.GroupSize();
			m_Points.resize(size);

			for (int i = 0; i < size; i++)
			{
			  m_Points[i] = in_Set[i];
			}

			return (*this);
		}

		/////////////////////////////////////////////////////////////////////////////////////
		int GroupSize() const
		{
		  return m_groupSize;
		}

    /////////////////////////////////////////////////////////////////////////////////////
		void InitGroupSize(int i)
		{
			if (m_Points.size() != 0)
			{
				throw "\nInitialization of the group size can be made only when the set is empty\n";
				return;
			}

			m_groupSize = i;

			m_MeanX.resize(m_groupSize);
			m_MeanY.resize(m_groupSize);
			m_MeanDev.resize(m_groupSize);

			ClearMeans();
		}
		
		/////////////////////////////////////////////////////////////////////////////////////
		int Size() const
		{
			return int(m_Points.size()) / m_groupSize;
		}
		
		/////////////////////////////////////////////////////////////////////////////////////
		void Clear()
		{
			m_Points.clear();
		}

		/////////////////////////////////////////////////////////////////////////////////////
		void Add(std::vector<Point> &in_p)
		{
			int size = in_p.size();

			if (size - size / m_groupSize * m_groupSize > 0)
			{
				throw "\nThe input group of pounts has not appropriate  size";
				return;
			}

			int prevSize = m_Points.size();

			m_Points.resize(prevSize + size);

			for (int i = 0; i < size; i++)
			{
				m_Points[prevSize + i] = in_p[i];
			}

			ClearMeans();
		}

		/////////////////////////////////////////////////////////////////////////////////////
		void Disable(T_Points2DSet<Point> &in_Set)
		{
			int size = in_Set.Size() * in_Set.GroupSize();

		  for (int i = 0; i < size; i++)
			{
				int numb = in_Set[i].Numb();
				if (numb >= 0)
				{
			    m_Points[numb].SetNumb(-1);
				}
			}
		}

		////////////////////////////////////////////////////////////////////////////////////
		void EnableAll()
		{
			for (int i = 0; i < Size(); i++)
			{
				for (int j = 0; j < GroupSize(); j++)
				{
					m_Points[GroupSize() * i + j].SetNumb(i);
				}
			}
		}

		/////////////////////////////////////////////////////////////////////////////////////
		void RemoveGroup(int in_i)
		{
			int s = Size();

			if (in_i < Size())
			{
				/*for (int j = in_i - 1; j < in_i + 3; j++)
				{
					std::cout << std::endl << j << std::endl
						<< "  x = " << m_Points[GroupSize() * j].X()
						<< "  y = " << m_Points[GroupSize() * j].Y()
						<< std::endl << "  x = " << m_Points[GroupSize() * j + 1].X()
						<< "  y = " << m_Points[GroupSize() * j + 1].Y();
				}*/

				for (int j = 0; j < GroupSize(); j++)
				{
					m_Points.erase(m_Points.begin() + GroupSize() * in_i);
					//std::cout << std::endl << "in_i = " << in_i << "  j = " << j;
				}

				/*for (int j = in_i - 1; j < in_i + 2; j++)
				{
					std::cout << std::endl << j << std::endl
						<< "  x = " << m_Points[GroupSize() * j].X()
						<< "  y = " << m_Points[GroupSize() * j].Y()
						<< std::endl << "  x = " << m_Points[GroupSize() * j + 1].X()
						<< "  y = " << m_Points[GroupSize() * j + 1].Y();
				}*/
			}
			else
			{
				std::cout << std::endl << "not deleted:  in_i = " << in_i << "  size = " << Size();
			}

			if (s - Size() != 1)
			{
				std::cout << std::endl << "s - size = " << s - Size();
			}
		}

		/////////////////////////////////////////////////////////////////////////////////////
		Point operator[] (int i) const
		{
			return m_Points[i];
		}
		
		/////////////////////////////////////////////////////////////////////////////////////
		void Load(const char *in_filename,
							const double in_t,
							const int in_imHeight)
		{
		std::ifstream inFile;
		inFile.open(in_filename, std::ios::in);

		if (!inFile)
		{
			std::cerr << "Can't open " << in_filename << "\n";
			return;
		}

		float x1, y1, x2, y2, c, dist, distPrev;
		int numb = 0;
		int n;
		inFile >> n;

		for (int i = 0; i < n; i++)
		{
			inFile >> c;
			inFile >> dist;
			inFile >> distPrev;
			inFile >> c;
			inFile >> c;

			inFile >> x1;
			inFile >> y1;

			for (int j=0; j<130; j++)
			{
				inFile >> c;
			}

			inFile >> x2;
			inFile >> y2;

			for (int j=0; j<130; j++)
			{
				inFile >> c;
			}
			
			//std::vector<C_Point2D> P;
			if (in_t * dist < distPrev)
			{
				//P.push_back(C_Point2D(x1, y1, 2 * numb));
				//P.push_back(C_Point2D(x2, y2, 2 * numb + 1));

				//out_PointsSet.Add(P);
				m_Points.push_back(C_Point2D(x1, in_imHeight - y1, 2 * numb));;
				m_Points.push_back(C_Point2D(x2, in_imHeight - y2, 2 * numb + 1));

				numb++;
			}
		}

		inFile.close();
	}
		
		/////////////////////////////////////////////////////////////////////////////////////
		// Save points related to one of the images
		// in_n - number of point in group to save =
		//        = number of image
		void Save(char *in_filenameX,
			        char *in_filenameY,
			        int in_n)
		{
			std::ofstream outFileX;
			std::ofstream outFileY;
			outFileX.open(in_filenameX, std::ios::out);
			outFileY.open(in_filenameY, std::ios::out);

			for (int i = 0; i < Size(); i++)
			{
				outFileX << m_Points[GroupSize() * i + in_n].X() << " ";
				outFileY << m_Points[GroupSize() * i + in_n].Y() << " ";
			}

			outFileX.close();
			outFileY.close();
		}

		/////////////////////////////////////////////////////////////////////////////////////
		// Moves center of mass of points to (0, 0) and sqrt(MaxX^2 + MaxY^2) -> sqrt(2)
		// Returns transformation matrices
		double Normalize(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out_T,
									   Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out_T1,
									   const double in_sigma)
	  {
			std::vector<double> coef(m_groupSize);

			for (int i = 0; i < m_groupSize; i++)
			{
				coef[i] = sqrt(2.0) / MeanDev(i);
			}

			for (int i = 0; i < Size(); i++)
			{
				for (int j = 0; j < m_groupSize; j++)
				{
					m_Points[m_groupSize * i + j].Shift(MeanX(j), MeanY(j));
					m_Points[m_groupSize * i + j].Scale(coef[j]);
				}
			}
						
			// Computing of normalization matrices
			C_TransformMatrices::Normalization(out_T, m_MeanX[0], m_MeanY[0], m_MeanDev[0],
				                                                                       in_sigma);
			double err = C_TransformMatrices::Normalization(out_T1, m_MeanX[1], m_MeanY[1],
				                                                         m_MeanDev[1], in_sigma);
			ClearMeans();
			return err;
		}

		/////////////////////////////////////////////////////////////////////////////////////
		// Moves origin from the up left corner to the center of the image
		// and sqrt(MaxX^2 + MaxY^2) -> sqrt(2)
		double Normalize(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out_T,
									   Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out_T1,
									   int in_width,
			               int in_height,
										 const double in_sigma)
	  {
			std::vector<double> coef(m_groupSize);

			std::cout << std::endl << std::endl;
			for (int i = 0; i < m_groupSize; i++)
			{
				coef[i] = sqrt(2.0) / MeanDev(i);
				std::cout << "coef[i] = " << coef[i] << "  ";
			}

			std::cout << std::endl << std::endl << "MeanX[0] - in_width / 2.0 = " <<
				MeanX(0) - in_width / 2.0 << " | MeanY[0] - in_heigth / 2.0 = " <<
				MeanY(0) - in_height / 2.0 << std::cout << "MeanX[1] - in_width / 2.0 = " <<
				MeanX(1) - in_width / 2.0 << " | MeanY[1] - in_heigth / 2.0 = " <<
				MeanY(1) - in_height / 2.0;

			for (int i = 0; i < Size(); i++)
			{
				for (int j = 0; j < m_groupSize; j++)
				{
					m_Points[m_groupSize * i + j].Shift(in_width / 2.0, in_height / 2.0);
					m_Points[m_groupSize * i + j].Scale(coef[j]);
				}
			}
					
			// Computing of normalization matrices
			C_TransformMatrices::Normalization(out_T, in_width / 2.0, in_height / 2.0,
				                                                         m_MeanDev[0], in_sigma);
			double err = C_TransformMatrices::Normalization(out_T1, in_width / 2.0,
				                                        in_height / 2.0, m_MeanDev[1], in_sigma);
			ClearMeans();
			return err;
		}
		
		/////////////////////////////////////////////////////////////////////////////////////
		void RetPoints (T_Points2DSet<Point> &out_P,
										int in_c) const
	  {
			// rand: 0 - 2^15-1  =>  0 - Size()
			double coef = (Size() - 1) / (pow(2.0, 15) - 1);
								
			for (int i = 0; i < in_c; i++)
			{
				int n = -1;
				int j;
				
				while (n < 0)
				{
				  j = (int)(rand() * coef);
					n = m_Points[2 * j].Numb();
				}
				
				std::vector<C_Point2D> P;
				P.push_back(m_Points[2 * j]);
				P.push_back(m_Points[2 * j + 1]);
				out_P.Add(P);
			}
		}

		/////////////////////////////////////////////////////////////////////////////////////
		// Makes a set of groups of n+1 corresponding points - *out,
		// *in_groups is a set of groups of n cor. points
		// *in_pairs is a set of pairs of cor. points
		// The last points in groups and the first points in pairs must be from the same image
		static void Combine(T_Points2DSet<Point> &out,
												const T_Points2DSet<Point> &in_groups,
												const T_Points2DSet<Point> &in_pairs,
												int n)
		{
			int numbGroups = in_groups.Size();
			int numbPairs = in_pairs.Size();

			T_Points2DSet<Point> temp(n + 1);
			temp.Clear();

			for (int i = 0; i < numbGroups; i++)
			{
				int numb1 = in_groups[i * n + n - 1].Numb();

				for (int j = 0; j < numbPairs; j++)
				{
					int numb2 = in_pairs[j * 2].Numb();
				  
					if (numb1 == numb2)
					{
						std::vector<C_Point2D> P;

						for (int k = 0; k < n; k++)
						{
							P.push_back(in_groups[i * n + k]);
						}
						P.push_back(in_pairs[j * 2 + 1]);

						temp.Add(P);
						break;
					}
					/*if ((p1.X() != p2.X() || p1.Y() != p2.Y()) && (p1.Numb() == p2.Numb()))
					{
						std::cout << std::endl << p1.X() << "  " << p2.X() << "     " << p1.Y() << "  "
							<< p2.Y() << "      " << p1.Numb();
					}*/
				}
			}

			out = temp;
		}
		
		/////////////////////////////////////////////////////////////////////////////////////
		void Print() const
		{
			std::cout << "\n";

			for (int i=0; i<(int)m_Points.size(); i++)
			{
				C_Point2D P = m_Points[i];
				std::cout << P.Numb() << "  " << P.X() << "  " << P.Y() <<"\n";
			}
		}

    /////////////////////////////////////////////////////////////////////////////////////
	};
}
#endif // POINTSSET_H_
