#ifndef HOUGH_TRANSFORM_H
#define HOUGH_TRANSFORM_H

// system includings
#define _USE_MATH_DEFINES
#include <cmath>

//own includings
#include "ModelTypes.h"
#include "PointsContainer.h"

namespace Vtl
{
namespace CurveDetection
{
  template <typename IterType, typename ModelType>
	void HoughTransform(std::vector<Point> &points, 
		                  ModelType &model,   
		                  IterType img,   //image where found pattern will be drawn
		                  int width,      //image
											int height,     //      size
											int nls,        // image next line shift
							        const char *fname,  // src image name which is used as
											const char *prefix) // prefix for all saved images
	{
		model.InitAccum();
   
		// fill accumulator
		int size = points.size();
		std::cout << std::endl << std::endl << "points size = " << size;

    for (int i = 0; i < size; i++)
		{
		  model.AddToAccum(&points[i]);
		}
    model.CalcParams();

		// save accumulator
		std::auto_ptr<AImage> _acc(::AImageCreate(0, AImage::IMGC_GRAYSCALE, 
																              model.sizeFi, model.sizeRo, 8));
		AImageMem8* acc = dynamic_cast<AImageMem8*>(_acc.get());
    model.SaveAccum(acc,fname,prefix);

		// select and save models
		model.SaveModels(img, acc, width, height, nls, fname,prefix);
	}

	/////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////
	template <typename T>
	class HoughFuncForLine
	{
	public:
		int sizeFi;
		int sizeRo;

	private:
		std::vector<ModelPointers <Line <T> > > accum;
		T stepFi;
		T stepRo;
		int max;
		T maxRo;

	public:
		HoughFuncForLine(T in_stepFi,
										 T in_stepRo,
										 int imgW,
										 int imgH)
		{
			stepFi = T( M_PI/180 * in_stepFi );
			stepRo = in_stepRo;

			sizeFi = int( 2 * M_PI / stepFi);
			maxRo = sqrt(T(imgW*imgW + imgH*imgH));
			sizeRo = int( maxRo / stepRo) + 1;

			max = -1;
		}
		//-------------------------------------------------------------------------
		HoughFuncForLine(int accumW,
			               int accumH,
										 int imgW,
										 int imgH)
		{
		  sizeFi = accumW;
			sizeRo = accumH;

			stepFi = T(2) * T(M_PI) / T(accumW);
      maxRo = sqrt(T(imgW*imgW + imgH*imgH));
			stepRo = maxRo / T(accumH);

			std::cout << std::endl << "sizeFi = " << sizeFi << "  sizeRo = " << sizeRo
				<< "  stepFi = " << 180 * stepFi / M_PI << "  stepRo = " << stepRo;

			max = -1;
		}

		//-------------------------------------------------------------------------
		void InitAccum()
		{
			max = -1;
			accum.clear();
		  accum.resize(sizeFi * sizeRo);
		}

		//-------------------------------------------------------------------------
		void AddToAccum(const Point *p)
		{
			T fi = T(-M_PI);
			int k = int(stepFi / (0.5 / 180 * M_PI)) + 1;
			T step = stepFi / T(k);
			
			for (float i = 0; i < sizeFi; i += 1.0f/k, fi += step)
			{
				//std::cout << std::endl << fi;
				T ro = Line<T>::Ro(p->x, p->y, fi);

				if (ro < 0)
				{
				  continue;
				}

				int j = int(ro / stepRo);
				int n = j * sizeFi + int(i);
				
				bool add = false;

				if (accum[n].points.size() == 0)
				{
				  add = true;
				}
				else if (p != accum[n].points[accum[n].points.size() - 1])
				{
					add = true;
				}

				if (add)
				{
				  accum[n].params.cosFi += T(cos(double(fi)));
					accum[n].params.sinFi += T(sin(double(fi)));
				  accum[n].params.ro += ro;
          accum[n].points.push_back(p);
				}
			}
		}

		//-------------------------------------------------------------------------
		void CalcParams()
		{
			int size = accum.size();
		  for (int i = 0; i < size; i++)
			{
				int n = accum[i].points.size();
				
				if (n > 0)
				{
					accum[i].params.ro /= n;
					accum[i].params.cosFi /= n;
					accum[i].params.sinFi /= n;

					/*T cosFi = accum[i].cosFi;
					T sinFi = accum[i].sinFi;
					T r = accum[i].ro;

					for (int j = accum[i].points.size() - 1; j >= 0; j--)
					{
						double x = accum[i].points[j]->x;
						double y = accum[i].points[j]->y;

					  if (fabs(x * cosFi + y * sinFi - r) > stepRo / T(2))
						{
							accum[i].points.erase(accum[i].points.begin() + j);
						}
					}

					if (n != accum[i].points.size() && n > 100)
					{
						std::cout << std::endl << n << "  " << accum[i].points.size();
					}*/
				}
			}
		}

		//-------------------------------------------------------------------------
		int FindMax()
		{
			if (max < 0)
			{
				if (accum.size() > 0)
				{
					int size = accum.size();

					for (int i = 0; i < size; i++)
					{
						if (int(accum[i].points.size()) > max)
						{
							max = accum[i].points.size();
						}
					}
				}
			}
			return max;
		}

		//-------------------------------------------------------------------------
		void SaveAccum(AImageMem8 *acc,const char *fname,const char *prefix)
		{
			float coef = float(FindMax());
			std::cout << std::endl << "FindMax = " << coef;
			coef = 255.0f / coef;
			std::cout << "  coef = " << coef;

			AImageMem8::Pixel *p = (*acc)[0];

			int i = 0;
			for (int y = 0; y < sizeRo; y++)
			{
				for (int x = 0; x < sizeFi; x++)
				{
					(*p) = AImageMem8::Pixel(accum[i].points.size() * coef);
					p++;
					i++;
				}
			}

			{
				std::ostringstream os;
				os << YFileSys::getNameWithoutExt(fname) << "_#_" << "HT_" << prefix << "_Sinogr" << ".tif";//YFileSys::getExt(base_name);
				::AImageSave(acc, os.str().c_str());
			}

		}

		//-------------------------------------------------------------------------
		bool IsMax(int iFi,
			         int iRo,
							 int vic)
		{
		  int v = accum[iRo * sizeFi + iFi].points.size();
		  int max = 0;

			if (v > FindMax() * 0.4)
			{
				int maxX = std::min(iFi + vic, sizeFi);
				int maxY = std::min(iRo + vic, sizeRo);

				int c = 0; 
				float summ = 0;

				for (int j = std::max(iRo - vic, 0); j < maxY; j++)
				{
					for (int i = std::max(iFi - vic, 0); i < maxX; i++)
					{
						if (i == iFi && j == iRo)
						{
						  continue;
						}

						int v2 = accum[j * sizeFi + i].points.size();

						if (v2 >= v)
						{
							return false;
						}
						summ += v2;
						c++;

						if (v2 > max)
						  if (!(i == iFi-2 && (j == iRo-2 || j == iRo-1 || j == iRo || j == iRo+1 || j == iRo+2)))
						    if (!(i == iFi-1 && (j == iRo-2 || j == iRo-1 || j == iRo || j == iRo+1 || j == iRo+2)))
							    if (!(i == iFi && (j == iRo-2 || j == iRo-1 || j == iRo || j == iRo+1 || j == iRo+2)))
							      if (!(i == iFi+1 && (j == iRo-2 || j == iRo-1 || j == iRo || j == iRo+1 || j == iRo+2)))
								      if (!(i == iFi+2 && (j == iRo-2 || j == iRo-1 || j == iRo || j == iRo+1 || j == iRo+2)))
						{  
						  max = v2;
						}
					}
				}

				vic = 2;

				for (int j = std::max(iRo - vic, 0); j < maxY; j++)
				{
					for (int i = std::max(iFi - vic, 0); i < maxX; i++)
					{
						int v2 = accum[j * sizeFi + i].points.size();
						summ -= v2;
						c--;
					}
				}

				summ /= float(c);

				std::cout << std::endl << "max x = " << iFi << "  y = " << iRo
						<< "  v = " << v << "  T = " << summ / v << "  max = " << max
            << "  max/v = " << float(max)/float(v);
				if (summ / v < 0.3)
				{
				  std::cout << "  !!!";
				  return true;
				}
			}
			return false;
		}
		
		//-------------------------------------------------------------------------
		template <typename IterType>
		void SaveModels(IterType back,
			              AImageMem8* sinogr,
										int w,
										int h,
										int nls,
										const char *fname,const char *prefix)
		{
			// create destination backgroung
			std::auto_ptr<AImage> _dst(::AImageCreate(0, AImage::IMGC_GRAYSCALE, w, h, 8));
			AImageMem8* dst = dynamic_cast<AImageMem8*>(_dst.get());

			AImageMem8::Pixel *p_d = (*dst)[0];

			for (int y = 0; y < h; y++)
			{
				for (int x = 0; x < w; x++)
				{
					(*p_d) = AImageMem8::Pixel(*back);
					back++;
					p_d++;
				}
			}
			//{
			//	std::ostringstream os;
			//	os << YFileSys::getNameWithoutExt(fname) << "_#_" << "HT_" << prefix << "_dst" << ".tif";//YFileSys::getExt(base_name);
			//	::AImageSave(dst, os.str().c_str());
			//}

			//

			// select real lines
			int count = 0;
			AImageMem8::Pixel *s = (*sinogr)[0];

			for (int iRo = 0; iRo < sizeRo; iRo++)
			  for (int iFi = 0; iFi < sizeFi; iFi++)
				  if (IsMax(iFi, iRo, 20))
					{
						// draw line
					  accum[iRo * sizeFi + iFi].params.Draw((*dst)[0], w, h, nls);

						// mark maximum at the sinogram
						*(s + iRo * sizeFi + iFi) = AImageMem8::Pixel(255);
						*(s + (iRo+1) * sizeFi + (iFi+1)) = AImageMem8::Pixel(255);
						*(s + (iRo+1) * sizeFi + iFi) = AImageMem8::Pixel(255);
						*(s + iRo * sizeFi + (iFi+1)) = AImageMem8::Pixel(255);

						count++;
					}

			std::cout << std::endl << "lines: " << count;

			{
				std::ostringstream os;
				os << YFileSys::getNameWithoutExt(fname) << "_#_" << "HT_" << prefix << "_Lines" << ".tif";//YFileSys::getExt(base_name);
				::AImageSave(dst, os.str().c_str());
			}
			{
				std::ostringstream os;
				os << YFileSys::getNameWithoutExt(fname) << "_#_" << "HT_" << prefix << "_SinMax" << ".tif";//YFileSys::getExt(base_name);
				::AImageSave(sinogr, os.str().c_str());
			}
		}

		//---------------------------------------------------------------------------
		/*void Ends(int *x1,
							int *y1,
							int *x2,
							int *y2,
							T Fi,
							T ro,
							int w,
							int h)
		{
			if (fabs(Fi) < 0.01 || fabs(Fi - 3.141592654) < 0.01)
			{
			  (*y1) = 0;
        (*x1) = int(Line<T, Point>::X(0, Fi, ro) + 0.5);
				(*y2) = h - 1;
				(*x2) = int(Line<T, Point>::X(h - 1, Fi, ro) + 0.5);
				return;
			}

			(*x1) = 0;
			(*y1) = int(Line<T, Point>::Y(*x1, Fi, ro) + 0.5);
			(*x2) = w - 1;
			(*y2) = int(Line<T, Point>::Y(*x2, Fi, ro) + 0.5);

			if ((*y1) > h - 1)
			{
				(*y1) = h - 1;
				(*x1) = int(Line<T, Point>::X(h - 1, Fi, ro) + 0.5);
			}

			if ((*y1) < 0)
			{
				(*y1) = 0;
				(*x1) = int(Line<T, Point>::X(0, Fi, ro) + 0.5);
			}

			if ((*y2) > h - 1)
			{
				(*y2) = h - 1;
				(*x2) = int(Line<T, Point>::X(h - 1, Fi, ro) + 0.5);
			}

			if ((*y2) < 0)
			{
				(*y2) = 0;
				(*x2) = int(Line<T, Point>::X(0, Fi, ro) + 0.5);
			}

			if ((*x1) > w - 1)
			{
				(*x1) = w - 1;
				(*y1) = int(Line<T, Point>::Y(w - 1, Fi, ro));
			}

			if ((*x1) < 0)
			{
				(*x1) = 0;
				(*y1) = int(Line<T, Point>::Y(0, Fi, ro));
			}

			if ((*x2) > w - 1)
			{
				(*x2) = w - 1;
				(*y2) = int(Line<T, Point>::Y(w - 1, Fi, ro));
			}

			if ((*x2) < 0)
			{
				(*x2) = 0;
				(*y2) = int(Line<T, Point>::Y(0, Fi, ro));
			}
		}

		//---------------------------------------------------------------------------
		template <typename IterType>
		void DrawLine(IterType iter,
									int nls,
									int x1,
									int y1,
									int x2,
									int y2)
		{
			typedef Vtl::ImAlg::Bres::Draw<AImageMem8::Pixel> DrawFunctor;
			typedef Vtl::ImAlg::Bres::SetPixelGray<AImageMem8::Pixel, DrawFunctor> SetPixelFunctor;

			AImageMem8::Pixel color_pen = AImageMem8::Pixel(255);
			DrawFunctor draw_functor(color_pen);
			SetPixelFunctor set_pix(iter, nls, draw_functor);
			Vtl::ImAlg::Bres::Line<SetPixelFunctor> line(set_pix);

			set_pix.Set0(x1, y1);
			line.Draw(x2 - x1, y2 - y1);
		}*/
	};
}
}

#endif // HOUGH_TRANSFORM_H