#ifndef KMEANS_H
#define KMEANS_H

//system includings
#include <vector>
#include <iostream>
#include <math.h>

// DatumType is needed to have member val, over which clusterization is to be
// carried out
template <typename T, typename DatumType>
class Cluster1D
{
public:
	T val;                          // mean for cluster
	std::vector<const DatumType *> data;  // data items belonging to cluster

public:
	Cluster1D()
	{
	  val = T(0);
	}

	//---------------------------------------------------------------------------
	Cluster1D(const DatumType &d)
	{
	  val = d.val;
	}

	//---------------------------------------------------------------------------
	Cluster1D &operator + (const DatumType &d)
	{
	  val += d.val;
		data.push_back(&d);
		return (*this);
	}
	
	//---------------------------------------------------------------------------
	void Divide()
	{
		T size = T(data.size());
	  if (size > 0)
		{
		  val /= size;
		}
	}

	//---------------------------------------------------------------------------
	// ItemType - type of data item or cluster
	template <typename ItemType>
	T Dist(const ItemType &d)
	{
	  return T(fabs(double(val - d.val)));
	}
	//---------------------------------------------------------------------------
	void Print()
	{
    std::cout << "v = "
      << Vtl::Curvature::CurvatureScale<T>::value_to_radius(val);
	}
};

///////////////////////////////////////////////////////////////////////////////
// DatumType is needed to have members x and y, over which clusterization
// is to be carried out
template <typename T, typename DatumType>
class Cluster2D
{
public:
	T x;														      // mean 
	T y;														      //      for cluster
	std::vector<const DatumType *> data;  // data items belonging to cluster

public:
	Cluster2D()
	{
	  x = y = T(0);
	}

	//---------------------------------------------------------------------------
	Cluster2D(const DatumType &d)
	{
	  x = d.x;
		y = d.y;
	}

	//---------------------------------------------------------------------------
	Cluster2D &operator + (const DatumType &d)
	{
	  x += d.x;
		y += d.y;
		data.push_back(&d);
		return (*this);
	}
	
	//---------------------------------------------------------------------------
	void Divide()
	{
		T size = T(data.size());
	  if (size > 0)
		{
		  x /= size;
			y /= size;
		}
	}

	//---------------------------------------------------------------------------
	// ItemType - type of data item or cluster
	template <typename ItemType>
	T Dist(const ItemType &d)
	{
	  T dx = x - d.x;
		T dy = y - d.y;
	  return T(sqrt(double(dx * dx + dy * dy)));
	}
	//---------------------------------------------------------------------------
	void Print()
	{
		std::cout << "x = " << x << "  y = " << y;
	}
};

///////////////////////////////////////////////////////////////////////////////
// data item needed for clusterization of points over their value
template <typename T>
class DatumR
{
public:
	int x;			// coordinates
	int y;			// of point
  int sgn;    // curvature direction
	T val;			// curvature radius
	  
public:
	DatumR(int in_x,
		     int in_y,
				 T in_v)
	{
	  x = in_x;
		y = in_y;
		val = fabs(in_v);
    sgn = int(in_v / val);
    //std::cout << std::endl << "in = " << in_v <<  "  val = " << val
    //  << "  sgn = " << sgn;
	}

  //---------------------------------------------------------------------------
  DatumR()
  {
    x = y = sgn = 0;
    val = T(0);
  }

  //---------------------------------------------------------------------------
  DatumR<T>& operator = (const DatumR<T> &d)
  {
    x = d.x;
    y = d.y;
    val = d.val;
    sgn = d.sgn;
    return *this;
  }
};

///////////////////////////////////////////////////////////////////////////////
// data item needed for clusterization of their localization
template <typename T>
class DatumXY
{
public:
	T x;			// coordinates
	T y;			// of point
	const DatumR<T> *p;
	  
public:
	DatumXY(T in_x,
		      T in_y,
				  const DatumR<T> *in_point)
          : p(in_point)
	{
	  x = in_x;
		y = in_y;
	}
};

///////////////////////////////////////////////////////////////////////////////
template <typename DatumType, typename ClusterType>
void KMeans(std::vector<DatumType> &data,
						std::vector<ClusterType> &means,
            std::vector<float> &dev,
						int k)
{
	// initialization of means with random data items
	int step = int((data.size() - 2 - k )/ float(k));
  //std::cout << std::endl << "data size: " <<  data.size() << "  k = " << k
  //  << "  step: " << step;

  std::cout << std::endl;
  means.clear();

  if (step < 1)
  {
    std::cout << std::endl << "too little step";
    std::cout << "  =>  empty means were saved";
    return;
  }

  //std::cout << std::endl << std::endl << "Initial Means:  ";
  for (int i = 0; i < k; i++)
	{
    //std::cout << "i = " << k + step * i << "  ";
		means.push_back(ClusterType(data[k + step * i]));
    //means[i].Print();
    //std::cout << " | ";
	}
	
	std::vector<ClusterType> newMeans(k);

	float stepDif = 1.0f; // difference between previous and current means
	int size = data.size();
	
  // while means are not stable
	while (stepDif > 0.5f)
  {
		// initialize new set of means with 0
    newMeans.assign(k, ClusterType());
		dev.assign(k, 0);
    
    // for each data item
	  for (int i = 0; i < size; i++)
		{
			// calculate distanses to each mean and determine the closest mean
			int iCl = 0;
			float minDist = means[0].Dist(data[i]);

		  for (int j = 1; j < k; j++)
			{
			  float d = means[j].Dist(data[i]);
				if (d < minDist)
				{
				  minDist = d;
					iCl = j;
				}
			}
      
			newMeans[iCl] = newMeans[iCl] + data[i];
      dev[iCl] += minDist;
    }

    stepDif = 0;
		for (int i = 0; i < k;  i++)
		{
			// calculate mean value of cluster
			newMeans[i].Divide();

      // calculate deviation in cluster
      if (newMeans[i].data.size() > 0)
      {
        dev[i] /= float(newMeans[i].data.size());
      }

			// calculate difference between previous and current cluster
			stepDif += newMeans[i].Dist(means[i]);
      			
			means[i] = newMeans[i];
		}

	}
}

///////////////////////////////////////////////////////////////////////////////
// Draw edge points with the mean color of the cluster they belong to
template <typename IterType, typename T>
void DrawRadCluster(Cluster1D< T, DatumR<T> > *cluster,
								    IterType img,
		                int nls)
{
	int size = cluster->data.size();
	for (int i = 0; i < size; i++)
	{
		*(img + cluster->data[i]->y * nls + cluster->data[i]->x) = cluster->val;
	}
}

///////////////////////////////////////////////////////////////////////////////
// Draw center points and corresponding edge points of the same cluster.
// Also draw mean center of cluster with black point
template <typename IterType, typename T>
void DrawCentCluster(Cluster2D< T, DatumXY<T> > *cluster,
                     IterType img,
                     int width,
                     int height,
									   int nls,
                     T color)
{
	// draw all cluster data
  int size = cluster->data.size();
  for (int i = 0; i < size; i++)
	{
    // draw center point
		int x = int(cluster->data[i]->x);
		int y = int(cluster->data[i]->y);

    if (x >=0 && x < width && y >= 0 && y < height)
    {
		  *(img + y*nls + x) = color;
    }

    // draw corresponding edge point
    x = int(cluster->data[i]->p->x);
		y = int(cluster->data[i]->p->y);
		*(img + y*nls + x) = color;
	}

  // draw cluster center
  int x = int(cluster->x + 0.5);
	int y = int(cluster->y + 0.5);

  if (x >=0 && x < width && y >= 0 && y < height)
  {
	  *(img + y*nls + x) = T(255);

    y = int(cluster->y);
    *(img + y*nls + x) = T(255);

    x = int(cluster->x);
    y = int(cluster->y + 0.5);
    *(img + y*nls + x) = T(255);

    y = int(cluster->y);
    *(img + y*nls + x) = T(255);
  }
}

#endif // KMEANS_H