#ifndef _YGEOMETR_H
#define _YGEOMETR_H
/////////////////////////////////////////////////////////////////////////////////
#include <math.h>
#include "YTypes.h"

const double NM_PI = ::atan2(0.,-1.);
const double NM_2PI=NM_PI*2.;
const double NM_4PI=NM_PI*4.;
const double NM_PI2=NM_PI/2.;
const double NM_PI4=NM_PI/4.;
const double NM_DEG2RAD=NM_PI/180.;
const double NM_RAD2DEG=180./NM_PI;

#define rad(deg) (NM_DEG2RAD*(deg))
#define deg(rad) (NM_RAD2DEG*(rad))

#include <ostream>

// Points

template <class T>
class YPoint
{
public:
	T x;
	T y;

	YPoint( T _x = T(0), T _y = T(0)) { x = _x; y = _y; }
	bool operator==(const YPoint& p) const { return ( (x==p.x) && (y==p.y) ); }
	bool operator!=(const YPoint& p) const { return ( (x!=p.x) || (y!=p.y) ); }
	void operator=( const YPoint& p) { x = p.x; y = p.y; }
	double distance( const YPoint& to) const
	{
		return ::sqrt((x-to.x)*(x-to.x)+(y-to.y)*(y-to.y));
	}
	YPoint nearest( const YPoint& first, const YPoint& second) const
	{
		return (distance(first)<distance(second))?first:second;
	}
};

template<class T>
std::ostream& operator<<( std::ostream& os, const YPoint<T>& p)
{
	os << "( " << p.x << ", " << p.y << " )";
	return os;
}

typedef YPoint<int>    YPointInt;
typedef YPoint<double> YPointDbl;

inline bool pointsEqual( 
	const YPoint<double>& lhs,
	const YPoint<double>& rhs,
	double e = 0.0001)
{
	return ABS(lhs.x-rhs.x) < e
		&& ABS(lhs.y-rhs.y) < e;
}


template <class T>
class YRect
{
public: //TopLeft & BottomRight -ALL INCLUDES IN RECTANGLE AREA
	YPoint<T> TopLeft;
	YPoint<T> BottomRight;

	YRect( T l = T(0), T t=T(0), T r=T(0), T b=T(0))
	{ init(l,t,r,b); }
	YRect( YPoint<T> top_left, YPoint<T> bottom_right)
	{ init(top_left.x,top_left.y,bottom_right.x,bottom_right.y); }
	bool operator==( const YRect<T>& r) const 
	{
		return (TopLeft == r.TopLeft) && (BottomRight == r.BottomRight);
	}
	bool operator!=( const YRect<T>& r) const
	{
		return !(*this == r);
	}
	bool isIn( const YPoint<T>& p) const
	{
		return(p.x>=TopLeft.x && p.x<=BottomRight.x && p.y>=TopLeft.y && p.y<=BottomRight.y);
	}
	T& top() { return TopLeft.y; }
	T& bottom() { return BottomRight.y; }
	T& left() { return TopLeft.x; }
	T& right() { return BottomRight.x; }
	const T& top() const { return TopLeft.y; }
	const T& bottom() const { return BottomRight.y; }
	const T& left() const { return TopLeft.x; }
	const T& right() const { return BottomRight.x; }
	T width() const { return BottomRight.x - TopLeft.x; }
	T height() const { return BottomRight.y - TopLeft.y; }

	bool intersects( const YRect<T>& rect) const
	{
		T _left = MAX(left(),rect.left());
		T _right = MIN(right(),rect.right());
		T _top = MAX(top(),rect.top());
		T _bottom = MIN(bottom(),rect.bottom());

		// <= because bottom/right point is included in rect
		if ( _left <= _right && _top <= _bottom )
		{
			return true;
		}
		return false;
	}

	YRect<T> intersection( const YRect<T>& rect) const
	{
		T _left = MAX(left(),rect.left());
		T _right = MIN(right(),rect.right());
		T _top = MAX(top(),rect.top());
		T _bottom = MIN(bottom(),rect.bottom());

		YRect<T> result;
		// <= because bottom/right point is included in rect
		if ( _left <= _right && _top <= _bottom )
			result = YRect<T>(_left,_top,_right,_bottom);
		else
			result = YRect<T>(T(0),T(0),T(0),T(0));
		return result;
	}

	void expand( const YPoint<T>& p)
	{
		TopLeft.x = MIN(TopLeft.x,p.x);
		TopLeft.y = MIN(TopLeft.y,p.y);
		BottomRight.x = MAX(BottomRight.x,p.x);
		BottomRight.y = MAX(BottomRight.y,p.y);
	}
	void expand( const YRect<T>& rect)
	{
		TopLeft.x = MIN(TopLeft.x,rect.left());
		TopLeft.y = MIN(TopLeft.y,rect.top());
		BottomRight.x = MAX(BottomRight.x,rect.right());
		BottomRight.y = MAX(BottomRight.y,rect.bottom());
	}
	void init( T l, T t, T r, T b)
	{
		TopLeft.x = MIN(l,r);
		TopLeft.y = MIN(t,b);
		BottomRight.x = MAX(r,l);
		BottomRight.y = MAX(b,t);
	}
	bool empty()
	{
		return (0 == width()*height());
	}
};

template<class T>
std::ostream& operator<<( std::ostream& os, const YRect<T>& r)
{
	os << "( " << r.TopLeft << ", " << r.BottomRight << " )";
	return os;
}

typedef YRect<int>    YRectInt;
typedef YRect<double> YRectDbl;

class LineFun
{
public:
	static bool  isInLine( const YPointInt& begin, const YPointInt& end, const YPointInt& what, bool strict=false);
	static bool  isInLineBres4( const YPointInt& begin, const YPointInt& end, const YPointInt& p, bool strict=false);
	static bool  areIntersecting( const YPointInt& begin1, const YPointInt& end1, const YPointInt& begin2, const YPointInt& end2, bool strict=false);
	static bool  areIntersecting( const YPoint<double>& begin1, const YPoint<double>& end1, const YPoint<double>& begin2, const YPoint<double>& end2);
	//strict==true - NOT include edge points
	//strict==false - include edge points
	static YPointInt getIntersectionPoint( const YPointInt& begin1, const YPointInt& end1, const YPointInt& begin2, const YPointInt& end2);
	static YPointInt getIntersectionPointBres4( const YPointInt& begin1, const YPointInt& end1, const YPointInt& begin2, const YPointInt& end2);
	static bool  onSameSide( const YPointInt& from, const YPointInt& to, const YPointInt& p1, const YPointInt& p2);
	static int   getIntersectionPointBres4Y( const YPointInt& begin, const YPointInt& end, int X);
	static int   getIntersectionPointBres4X( const YPointInt& begin, const YPointInt& end, int Y);
};
class PolyFun
{
public:
	static bool isPointInPolygon(
		const YPoint<double>& point,
		const YPoint<double>* polygon, int polygon_points_number);
};

double distance( 
	const double x1, const double y1,
	const double x2, const double y2);

double PointOnAxisProjectionLength(
	const double axis_begin_x, const double axis_begin_y,
	const double axis_end_x, const double axis_end_y,
	const double point_x, const double point_y);

double TriangleArea( 
	const YPoint<double>& a,
	const YPoint<double>& b,
	const YPoint<double>& c);

double QuadrangleArea(
	const YPoint<double>& a,
	const YPoint<double>& b,
	const YPoint<double>& c,
	const YPoint<double>& d);

/////////////////////////////////////////////////////////////////////////////////
#endif //_YGEOMETR_H
