#ifndef _IMRESAMP_H
#define _IMRESAMP_H
////////////////////////////////////////////////////////////////////////////////////////////////////
#include "aimage.h"
////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////
// Class ImageResample provides such operations on images as rotate and resize.
// To use this class, you should create an object of this class (imod in the example),
// call init() function, then create output image using parameters obtained by
// functions width(), height(), bits() and imageClass(), and then call init() function
// with output image as argument.
//
// Example:
// 
//  AImage* from = ::AImageLoad( 0, true, "in.tif");
//  if(!from)
// 	  return 1;  //can't load image
////  ImageResample imod;
//  if(!imod.initResize( from, ImageResample::imBicubical, 1000, 1000)) // resize to 1000X1000
//	  return 1;
//  AImage* to = ::AImageCreate( 0, imod.imageClass(), imod.width(), imod.height(), imod.bits());
//				// output image created
//  if(!to)
// 	  return 1;
//  if(!imod.run( to))
//  	return 1;
//  ::AImageSave( to, "out.tif");
//
//  Class is currently working with grayscale and indexed color images.
//  Only nearest and histmax interpolation methods can work with indexed images.
//
//  Recommendations.
//
//  Bicubical method proved itself to be the best for rotating and enlarging of grayscale
//  images. Average method is best for shrinking.
//
//  ProgressCounter can be used if necessary. run() function accepts it as the second argument.
//  When ProgressCounter is avaliable run() can return false when a process
//  is interrupted by user.
////////////////////////////////////////////////////////////////////////////////////////////////////

class ImageResample
{
public:
	enum ModifyType
	{
		mtRotate,
		mtResize
	};
	enum InterpolationMethod
	{
		imUnknown,
		imAuto, // 
		imNearest,
		imBilinear,
		imBicubical,
		imAverage,
		imHistMax
	};
	struct ImageResampleParams
	{
		// ProgressCounter
		ProgressCounter* indicator;
		// input parameters
		ModifyType mtype;
		InterpolationMethod imeth;
		double radians; // angle of rotation in radians
		int new_width;
		int new_height;
		bool fill; // flag which determines should the empty spaces be filled
				   // in the proccess of rotation or shouldn't
		int fill_int; // intensity of the filling color
		// input image
		const AImage* from;
		// output image required parameters
		int to_width;
		int to_height;
		AImage::YImageClass to_class;
		int to_bits;
		// calculating parameters
		int shiftx, shifty; // parameters for rotated image
		// validity
		bool is_valid;
	};
private:
	ImageResampleParams prms;
	void nullifyPrms();
	bool init( ImageResampleParams _prms );
	void nullify() { nullifyPrms(); }
public:
	ImageResample() { nullify(); }
	virtual ~ImageResample() {}
	// Initialization function for resize, returns false if parameters are wrong
	// if one of parameters 'new_width' 'new_height' is <=0, it calculated to save aspect ratio
	bool initResize( const AImage* input_image, InterpolationMethod method, int new_width, int new_height=0);
	// Initialization function for rotate, returns false if parameters are wrong
	// if parameter 'radians' is false, parameter angle is in degrees
	bool initRotate( const AImage* input_image, InterpolationMethod method, double angle, bool radians=true, bool fill_empty_space = true, int fill_intensity = 0);
	// Use these functions to obtain parameters for output image
	virtual int Width() const { return prms.to_width; }
	virtual int Height() const { return prms.to_height; }
	virtual AImage::YImageClass ImageClass() const { return prms.to_class; }
	virtual int Bits() const { return prms.to_bits; }
	// output_image must be pre-created and must have parameters obtained by functions
	// declared above, returns false if process is interrupted
	virtual bool Run( AImage* output_image, ProgressCounter* indicator = 0);
};


// ImageRotate and ImageResize are separate classes for rotating and resizing
// respectively, both classes are derived from ImageResample and their usage is
// similar to it. 

class ImageRotate
: public ImageResample
{
public:
	ImageRotate() {}
	virtual ~ImageRotate() {}
	// Initialization function for rotate, returns false if parameters are wrong
	// if parameter 'radians' is false, parameter angle is in degrees
	bool init( const AImage* input_image, InterpolationMethod method, double angle, bool radians=true, bool fill_empty_space = true, int fill_intensity = 0)
	{
		return ImageResample::initRotate( input_image, method, angle, radians, fill_empty_space, fill_intensity);
	}
	// Use these functions to obtain parameters for output image
	virtual int Width() const { return ImageResample::Width(); }
	virtual int Height() const { return ImageResample::Height(); }
	virtual AImage::YImageClass ImageClass() const { return ImageResample::ImageClass(); }
	virtual int Bits() const { return ImageResample::Bits(); }
	// output_image must be pre-created and must have parameters obtained by functions
	// declared above, returns false if process is interrupted
	virtual bool Run( AImage* output_image, ProgressCounter* indicator = 0)
	{
		return ImageResample::Run( output_image, indicator);
	}
};

class ImageResize
: public ImageResample
{
public:
	ImageResize() {}
	virtual ~ImageResize() {}
	// Initialization function for rotate, returns false if parameters are wrong
	// if parameter 'radians' is false, parameter angle is in degrees
	bool init( const AImage* input_image, InterpolationMethod method, int new_width, int new_height=0)
	{
		return ImageResample::initResize( input_image, method, new_width, new_height);
	}
	// Use these functions to obtain parameters for output image
	virtual int Width() const { return ImageResample::Width(); }
	virtual int Height() const { return ImageResample::Height(); }
	virtual AImage::YImageClass ImageClass() const { return ImageResample::ImageClass(); }
	virtual int Bits() const { return ImageResample::Bits(); }
	// output_image must be pre-created and must have parameters obtained by functions
	// declared above, returns false if process is interrupted
	virtual bool Run( AImage* output_image, ProgressCounter* indicator = 0)
	{
		return ImageResample::Run( output_image, indicator);
	}
};



#endif // __IMRESAMP_H__
