
#ifndef IMTRANSF_H_
#define IMTRANSF_H_

#include <memory>
#include <vector>

#include "AImage.h"
#include "ygeometr.h"

class ProgressCounter;

class ATransformFunction
{
public:
	virtual YPoint<double> forward( const YPoint<double>& p) const = 0;
	virtual YPoint<double> reverse( const YPoint<double>& p) const = 0;
	virtual bool canForward( const YPoint<double>& p) const = 0;
	virtual bool canReverse( const YPoint<double>& p) const = 0;

	virtual ATransformFunction* clone() const = 0;
};

class CoordArray2D;
template<class Base, class Bit8, class Bit16> class ImageVectorT;
typedef ImageVectorT<AImage,AImageMem8,AImageMem16> ImageVector;
typedef ImageVectorT<const AImage,const AImageMem8,const AImageMem16> ImageVectorConst;

class ImageProxyVector;

class ImageTransformImpl;

class ImageTransform
{
public:
	enum EdgeMethod
	{
		None
	};
public:
	explicit ImageTransform( bool fastAlgorithm = true );
	virtual ~ImageTransform();

	// algorithm configuration options
	void setBackgroundColor( bool doCorrect, int value = 0);
	void setErrorColor( bool doCorrect, int value = 0);
	void setFlipHorizontal( bool value );
	void setFlipVertical( bool value );
	void setGridAdjustment( const YPoint<double>& gridNode);
//	void setHorizontalEdgesMethod( bool skip);
//	void setVerticalEdgesMethod( bool skip);

	bool init(
		const ImageProxyVector* inputVector,
		const YRect<int>& inputFragment,
		const CoordArray2D* coords,
		const ATransformFunction& transformFunction,
		double scale,
		ProgressCounter* indicator = 0,
		const EdgeMethod& edgeMethod = None);
	bool init(
		const ImageProxyVector* inputVector,
		const CoordArray2D* coords,
		const ATransformFunction& transformFunction,
		double scale,
		ProgressCounter* indicator = 0,
		const EdgeMethod& edgeMethod = None);
	int width() const;
	int height() const;
	std::vector<int> bits() const;
	std::vector<AImage::YImageClass> imageClass() const;
	YPoint<double> base() const;
	bool run( 
		ImageProxyVector* outputVector, 
		ProgressCounter* indicator = 0);
	bool run( 
		ImageProxyVector* outputVector, 
		const YRect<int>& outputFragment,
		ProgressCounter* indicator = 0);
private:
	// forbidden
	ImageTransform( const ImageTransform&);
	ImageTransform& operator=( const ImageTransform&);

	ImageTransformImpl* pImpl_;
};

class CoordArray2D
{
public:
	CoordArray2D( int width = 0, int height = 0)
	{ clear(); setSize(width,height); }

	void clear() { p_.clear(); w_ = 0; h_ = 0; }
	void setSize( int width, int height)
	{ w_ = width; h_ = height; p_.resize(w_*h_); }

	int width() const { return w_; }
	int height() const { return h_; }

	const YPoint<double>& at( int x, int y) const
	{ return p_[y*w_+x]; }

	const YPoint<double>* begin() const
	{ return &p_[0]; }
	const YPoint<double>* end() const
	{ return &p_[w_*h_]; }
	YPoint<double>* begin()
	{ return &p_[0]; }
	YPoint<double>* end()
	{ return &p_[w_*h_]; }
private:
	std::vector<YPoint<double> > p_;
	int w_;
	int h_;
};

bool operator==( const CoordArray2D& lhs, const CoordArray2D& rhs);
bool operator!=( const CoordArray2D& lhs, const CoordArray2D& rhs);
std::ostream& operator<<( std::ostream& os, const CoordArray2D& src);

template<class Base, class Bit8, class Bit16>
class ImageVectorT
{
public:
	typedef std::vector< std::pair<Bit8*,int> > Bit8Vector;
	typedef std::vector< std::pair<Bit16*,int> > Bit16Vector;
	typedef typename Bit8Vector::iterator iterator8;
	typedef typename Bit16Vector::iterator iterator16;
public:
	ImageVectorT() {}

	void add( Base* image);
	size_t planesCount8() const { return vec8_.size(); }
	size_t planesCount16() const { return vec16_.size(); }
	iterator8 planesBegin8() { return vec8_.begin(); }
	iterator8 planesEnd8() { return vec8_.end(); }
	iterator16 planesBegin16() { return vec16_.begin(); }
	iterator16 planesEnd16() { return vec16_.end(); }
	int width() const { return width_; }
	int height() const { return height_; }
	std::vector<int> bits() const { return bits_; }
	std::vector<typename Base::YImageClass> imageClass() const { return imageClass_; }
private:
	ImageVectorT( const ImageVectorT&);
	ImageVectorT& operator=( const ImageVectorT&);
private:
	std::vector<int> bits_;
	std::vector<typename Base::YImageClass> imageClass_;
	int width_;
	int height_;
	Bit8Vector vec8_;
	Bit16Vector vec16_;
};

template<class Base, class Bit8, class Bit16>
void ImageVectorT<Base,Bit8,Bit16>::add( Base* image)
{
	if(bits_.empty())
	{
		width_ = image->Width();
		height_ = image->Height();
	}
	if( image->Width() != width_
	 || image->Height() != height_ )
		throw "ImageVectorT::add(): Images should have same size.";
	bits_.push_back(image->Bits());
	imageClass_.push_back(image->ImageClass());
	for( int i = 0; i < image->NumPlanes(); ++i)
	{
		if( Bit8* bit8 = dynamic_cast<Bit8*>(image->GetPlane(i)) )
			vec8_.push_back(std::make_pair(bit8,(1<<bit8->Bits())-1));
		else if( Bit16* bit16 = dynamic_cast<Bit16*>(image->GetPlane(i)) )
			vec16_.push_back(std::make_pair(bit16,(1<<bit16->Bits())-1));
	}
}

class AImageProxy
{
public:
	virtual ~AImageProxy() {}

	virtual int width() const = 0;
	virtual int height() const = 0;
	virtual int bits() const = 0;
	virtual AImage::YImageClass imageClass() const = 0;

	virtual int get( int x, int y) const = 0;
	virtual int maxValue() const = 0;
	virtual void set( int x, int y, int value) = 0;

	virtual AImageProxy* clone() const = 0;
};

template<class ImageT>
class ImageProxy: public AImageProxy
{
public:
	ImageProxy( ImageT* image):
	image_(image), max_((1<<image->Bits())-1) {}
	ImageProxy( const ImageProxy& src)
		: image_(src.image_), max_(src.max_) {}
	virtual ~ImageProxy() {}

	virtual int width() const { return image_->Width(); }
	virtual int height() const { return image_->Height(); }
	virtual int bits() const { return image_->Bits(); }
	virtual AImage::YImageClass imageClass() const
	{ return image_->ImageClass(); }

	virtual int get( int x, int y) const
	{ return (*image_)[y][x]; }
	virtual int maxValue() const { return max_; }
	virtual void set( int x, int y, int value)
	{ (*image_)[y][x] = value; }
	AImageProxy* clone() const { return new ImageProxy(*this); }
private:
	ImageT* image_;
	const int max_;
};

template<class ImageT>
class ImageProxyConst: public AImageProxy
{
public:
	ImageProxyConst( const ImageT* image):
	image_(image), max_((1<<image->Bits())-1) {}
	ImageProxyConst( const ImageProxyConst& src)
		: image_(src.image_), max_(src.max_) {}
	virtual ~ImageProxyConst() {}

	virtual int width() const { return image_->Width(); }
	virtual int height() const { return image_->Height(); }
	virtual int bits() const { return image_->Bits(); }
	virtual AImage::YImageClass imageClass() const
	{ return image_->ImageClass(); }

	virtual int get( int x, int y) const
	{ return (*image_)[y][x]; }
	virtual int maxValue() const { return max_; }
	virtual void set( int x, int y, int value)
	{ throw "ImageProxyConst::set()"; }
	AImageProxy* clone() const { return new ImageProxyConst(*this); }
private:
	const ImageT* image_;
	const int max_;
};

class ImageProxyVector
{
public:
	typedef std::vector<AImageProxy*>::iterator iterator;
	typedef std::vector<AImageProxy*>::const_iterator const_iterator;
public:
	ImageProxyVector(): width_(0), height_(0) {}
	ImageProxyVector( const AImage* image)
	{ add(image); }
	ImageProxyVector( AImage* image)
	{ add(image); }
	virtual ~ImageProxyVector();
	ImageProxyVector( const ImageProxyVector& src);
	ImageProxyVector& operator=( const ImageProxyVector& src);

	void add( AImage* image);
	void add( const AImage* image);
	void add( const AImageProxy& imageProxy);
	size_t countImage() const { return bits_.size(); }
	size_t countProxy() const { return proxies_.size(); }
	iterator begin() { return proxies_.begin(); }
	iterator end() { return proxies_.end(); }
	const_iterator begin() const { return proxies_.begin(); }
	const_iterator end() const { return proxies_.end(); }
	int width() const { return width_; }
	int height() const { return height_; }
	std::vector<int> bits() const { return bits_; }
	std::vector<AImage::YImageClass> imageClass() const { return imageClass_; }
private:
	std::vector<AImageProxy*> proxies_;
	std::vector<int> bits_;
	std::vector<AImage::YImageClass> imageClass_;
	int width_;
	int height_;
};

void AImageSetNull( AImage* image);
void AImageSetColor( AImage* image, int r = 0, int g = 0, int b = 0);


#endif //IMTRANSF_H_
