#ifndef _NUM_CHOLESKY_H
#define _NUM_CHOLESKY_H
////////////////////////////////////////////////////////////////////////////////////////////////
namespace Vtl
{
namespace NumAlg
{
namespace Lin
{ 
////////////////////////////////////////////////////////////////////////////////////////////////

/** 
   <P>
   For a symmetric, positive definite matrix A, this function
   computes the Cholesky factorization, i.e. it computes a lower 
   triangular matrix L such that A = L*L'.
   If the matrix is not symmetric or positive definite, the function
   computes only a partial decomposition.  This can be tested with
   the is_spd() flag.

   <p>Typical usage looks like:
   <pre>
		ublas::c_matrix<float,3,3> A;
		ublas::c_matrix<float,3,3> L;
	 ... 
		bool isOk = cholesky::decomp(A,L);
		if(!isOk )
			throw  "factorization was not complete.";

	</pre>


   <p>
	(Adapted from JAMA, a Java Matrix Library, developed by jointly 
	by the Mathworks and NIST; see  http://math.nist.gov/javanumerics/jama).

   */
	class cholesky
	{
	public:
		/*!
		 *	Constructs a lower triangular matrix L, such that L*L'= A.
		 *	If A is not symmetric positive-definite (SPD), only a
		 *	partial factorization is performed.  If is_spd()
		 *	evalutate true (1) then the factorizaiton was successful.
		 ********************************************************************/
		template<class Matrix>
		static bool decomp(const Matrix &A,Matrix &L_)
		{
			typedef Matrix::value_type Real;

   			int m = A.size1();
			int n = A.size2();
			
			bool isspd = (m == n);

			if (m != n)
				throw "Cholseky decomposition require square matrix!";

			  // Main loop.
			 for (int j = 0; j < n; j++) 
			 {
				Real d(0);
				for (int k = 0; k < j; k++) 
				{
					Real s(0);
					for (int i = 0; i < k; i++) 
					{
					   s += L_(k,i)*L_(j,i);
					}
					L_(j,k) = s = (A(j,k) - s)/L_(k,k);
					d = d + s*s;
					isspd = isspd && (A(k,j) == A(j,k)); 
				 }
				 d = A(j,j) - d;
				 isspd = isspd && (d > 0);
				 L_(j,j) = std::sqrt(d > 0 ? d : Real(0) );
				 for (int k = j+1; k < n; k++) 
					L_(j,k) = 0;

			}
			 return isspd ;
		}
		/**
			Solve a linear system A*x = b, using the previously computed
			cholesky factorization of A: L*L'.

			\param  L   A Matrix with as many rows as A and any number of columns.
			\param  x   x so that L*L'*x = b.  If b is nonconformat, or if A
						was not symmetric posidtive definite, a null (0x0)
						array is returned.

			As a X argument first right side transfered. It will be replaced by result.
		*/
		template<class Matrix,class Vector>
		static void solve(Matrix &L,const Vector &x)
		{
			int n = L_.size1();
			if (x.size() != n)
				throw "cholesky::solve: Matrix and vectors dimansions are different";
			// Solve L*y = b;
			for (int k = 0; k < n; k++) 
			{
				for (int i = 0; i < k; i++) 
					x[k] -= x[i]*L(k,i);
				x[k] /= L(k,k);
			}
			// Solve L'*X = Y;
			for (int k = n-1; k >= 0; k--) 
			{
				for (int i = k+1; i < n; i++) 
					x[k] -= x[i]*L(i,k);
				x[k] /= L(k,k);
			}
		}
		/**
		 *  Solve a linear system A*X = B, using the previously computed
		 *  cholesky factorization of A: L*L'.
		 *  
		 *  \param  L   A Matrix with as many rows as A and any number of columns.
		 *  \param  X   so that L*L'*X = B.  If B is nonconformat, or if A
		 *  was not symmetric posidtive definite, a null (0x0)
		 *  array is returned.
		 ***************************************************************************/
		template<class Matrix,class Matrix2>
		static void solve(const Matrix& L,const Matrix2 &X)
		{
			int n = L.size1();
			if (x.size1() != n)
				throw "";

			int nx = X.size2();

			// Solve L*y = b;
			for (int j=0; j< nx; j++)
			{
				for (int k = 0; k < n; k++) 
				{
					for (int i = 0; i < k; i++) 
						X(k,j) -= X(i,j)*L(k,i);
					X(k,j) /= L(k,k);
				}
			}
			// Solve L'*X = Y;
			for (int j=0; j<nx; j++)
			{
				for (int k = n-1; k >= 0; k--) 
				{
					for (int i = k+1; i < n; i++) 
						X(k,j) -= X(i,j)*L(i,k);
					X(k,j) /= L(k,k);
				}
			}
		}
	};
////////////////////////////////////////////////////////////////////////////////////////////////
}//namespace Lin
}//namespace NumAlg
}//namespace Vtl
////////////////////////////////////////////////////////////////////////////////////////////////
#endif //_NUM_CHOLESKY_H
