#ifndef _PL_FILTER_ELEM_H
#define _PL_FILTER_ELEM_H
/////////////////////////////////////////////////////////////////////////////
#include <iostream>
#include <algorithm>
#include "plf_port.h"
namespace Vtl
{
namespace Plf 
{
/** @addtogroup PipeLineFiltersDoxGroup  Pipe Line Filters. 
 *  @{
 */
/////////////////////////////////////////////////////////////////////////////////
/// class FilterBase
/// @brief  .
/////////////////////////////////////////////////////////////////////////////////
	class FilterBase
	{
		std::string _name;
		const int num_inp;
		const int num_out;
		std::vector<InPortBase*>  in_ports;
		std::vector<OutPortBase*> out_ports;
	private:
		FilterBase& operator=(const FilterBase&); //!>forbidden
		FilterBase(const FilterBase&); //!>forbidden
	protected:
		FilterBase(int num_in_ports,int num_out_ports)
			:num_inp(num_in_ports),num_out(num_out_ports)
		{
			in_ports.resize(num_inp);
			out_ports.resize(num_out);
		}
		FilterBase(std::string name,int num_in_ports,int num_out_ports)
			:_name(name),num_inp(num_in_ports),num_out(num_out_ports)
		{
			in_ports.resize(num_inp);
			out_ports.resize(num_out);
		}
	public:
		virtual ~FilterBase()
		{
		}
		void setName(std::string name) {_name=name;}
		virtual std::string Name() const {return _name;}
		virtual int NumInputs() const {return num_inp;}
		virtual int NumOutputs() const {return num_out;}
		virtual InPortBase* getInPort(int i) const 
		{
			assert(i>=0 && i<NumInputs());
			return in_ports[i];
		}
		virtual OutPortBase* getOutPort(int i) const 
		{
			assert(i>=0 && i<NumOutputs());
			return out_ports[i];
		}
		virtual bool isComplex() const=0;
		virtual void AddPort(InPortBase& port,int i)
		{
			in_ports[i]=&port;
		}
		virtual void AddPort(OutPortBase& port,int i)
		{
			out_ports[i]=&port;
		}
	};
/////////////////////////////////////////////////////////////////////////////////
/// class FilterComplex
/// @brief      () ,
///     /  .
///      class FilterElementary
///
///    ,    ,
///	  .      
///       .   
///	  .         
///	   .   -    ,
///       ,     
///	  . \n
/// \b :       , , 
///         
///	  ,       .   
///	       : ,
/// ,  ,        
///    C#.
/////////////////////////////////////////////////////////////////////////////////
	class FilterComplex
		: public FilterBase
	{
	private:
		FilterComplex& operator=(const FilterComplex&); //!>forbidden
		FilterComplex(const FilterComplex&); //!>forbidden
	protected:
		FilterComplex(int num_in_ports,int num_out_ports)
			:FilterBase(num_in_ports,num_out_ports)
		{
		}
		FilterComplex(std::string name,int num_in_ports,int num_out_ports)
			:FilterBase(name,num_in_ports,num_out_ports)
		{
		}
	public:
		virtual ~FilterComplex()
		{
		}
		virtual bool isComplex() const { return true;}
	};
/////////////////////////////////////////////////////////////////////////////////
/// class FilterElementary
/// @brief       .
/////////////////////////////////////////////////////////////////////////////////
	class FilterElementary
		: public FilterBase
	{
	private:
		bool in_queue;
		bool finished;
		int h_max;
		int y_pos;
	public:
		int id;
	private:
		FilterElementary& operator=(const FilterElementary&); //!>forbidden
		FilterElementary(const FilterElementary&); //!>forbidden
	protected:
		FilterElementary(int num_in_ports,int num_out_ports)
			:FilterBase(num_in_ports,num_out_ports)
		{
		}
		FilterElementary(std::string name,int num_in_ports,int num_out_ports)
			:FilterBase(name,num_in_ports,num_out_ports)
		{
		}
	public:
		virtual ~FilterElementary()
		{
		}
		virtual bool isComplex() const { return false;}
		virtual void AddPort(InPortBase& port,int i,Extent ext)
		{
			port.Init(this,i,ext);
			FilterBase::AddPort(port,i);
		}
		virtual void AddPort(OutPortBase& port,int i)
		{
			port.Init(this,i);
			FilterBase::AddPort(port,i);
		}
		virtual void processLine()=0;
		virtual bool isFinished() const
		{
			return finished;
		}
		virtual void set_OutPortsImagesSizeAndInPortsExtents()
		{
			ImgSize ims=getInPort(0)->getImageSize();
			//   :
			//    -   
			for(int i=1; i<NumInputs() ; i++)
			{
				ImgSize  s=getInPort(i)->getImageSize();
				if(s!=ims)
				{
					std::ostringstream os;
					os << "Filter '" << this->Name() << "' image sizes in InPorts " << i-1 << " and " << i 
						<< " is different: " << ims << " vs " << s ;
					throw os.str();
				}
			}
			//       
			for(int i=0; i<NumOutputs() ; i++)
			{
				getOutPort(i)->set_bufferImgSize(ims);
			}
			// Extension -    ,    .
		}
		void init_OutPortsExtents()
		{
			for(int i=0; i<NumOutputs() ; i++)
			{
				getOutPort(i)->init_bufferExtents();
			}
		}
		virtual void set_total_delay()
		{
			int delay=0;
			for(int i=0; i<NumInputs() ; i++)
				delay=std::max(delay,getInPort(i)->total_delay);
			for(int i=0; i<NumOutputs() ; i++)
				getOutPort(i)->set_total_delay(delay);
		}

		void init_buffers()
		{
			h_max=0;
			y_pos=0;
			for(int i=0; i<NumOutputs() ; i++)
			{
				getOutPort(i)->init_buffer();
				h_max=std::max(getOutPort(i)->getImageSize().height,h_max);
			}
			in_queue=false;
			finished=false;
		}
		void put_in_queue(Fifo<int>& Q)
		{
			if(!in_queue)
			{
				in_queue=true;
				Q.put(this->id);
			}
		}
		virtual void TryPutInQueue(Fifo<int>& Q)
		{
			bool ready=true;;
			for(int i=0; i<NumOutputs(); ++i)
			{
				OutPortBase* out=this->getOutPort(i);
				if(out->inPopTailStage())
				{
					//printf("PopTail %s\n",this->Name().c_str());
					out->PopTail();
				}
				else
					ready = ready & out->isReadyForUpdate();
			}
			if(!isFinished())
			{
				for(int i=0; i<NumInputs(); ++i)
				{
					InPortBase* in=this->getInPort(i);
					ready = ready & in->isDataReady();
				}
				if(ready)
				{
///exparimental
					bool any_neighbour_in_queue=false;
					for(int i=0; i<NumOutputs(); ++i)
					{
						OutPortBase* out=this->getOutPort(i);
						any_neighbour_in_queue=any_neighbour_in_queue || out->getFilter()->in_queue;
					}
					for(int i=0; i<NumInputs(); ++i)
					{
						InPortBase* in=this->getInPort(i);
						any_neighbour_in_queue=any_neighbour_in_queue || in->getFilter()->in_queue;
					}
					if(!any_neighbour_in_queue)
///end exparimental
					put_in_queue(Q);
				}
			}
		}
		///    .   
		///    1    processLine()
		///   0        .
		///         ( 
		///  )        .
		///   ,        -
		///       ()   
		///       ,  :
		///     y_tmp_in =[y_gen/(Hmax-1)*(InP.height-1)]; \n
		///     y_tmp_in <= InP.y_cur <= y_tmp_in +1 .\n
		///         
		///     y_tmp_in == InP.y_cur ,     .\n
		///      ,    \n
		///     y_tmp_out=[y_gen/(Hmax-1)*(OutP.height-1)]; \n
		///     y_tmp_out <= OutP.y_wait <= y_tmp_out+1 .\n
		///         
		///     y_tmp_out == OutP.y_wait ,     .\n
		///  y_tmp_in -     ,  y_tmp_out -   .
		/// ..       .
		///         InP,  ,
		///           .
		///       - exception.
		int Y_pos() const { return y_pos;}
		void do_1_step()
		{
			in_queue=false;
			processLine();
			++y_pos;
			if(y_pos==h_max)
			{
				Finish();
			}
		}
		void Finish()
		{
			for(int i=0; i<NumInputs(); ++i)
			{
				InPortBase* in=this->getInPort(i);
				in->Finish();
			}
			for(int i=0; i<NumOutputs(); ++i)
			{
				OutPortBase* out=this->getOutPort(i);
				out->Finish();
			}
			finished=true;
		}
	};
/** @} */ // end of PipeLineFiltersDoxGroup
} //namespace Plf
} //namespace Vtl
#endif //_PL_FILTER_ELEM_H
